__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os
import shlex
import time

from tunel.logger import logger

from .container import ContainerLauncher


class Singularity(ContainerLauncher):
    name = "singularity"

    def run(self, *args, **kwargs):
        """
        Run handles some command to singularity (e.g., run or exec)
        """
        command = "%s %s singularity %s" % (self.path, self.environ, " ".join(args[0]))

        # If the user wants a shell, give to them!
        if "shell" in args[0]:
            self.ssh.shell(command, interactive=True)
        else:
            res = self.ssh.execute(command)
            self.ssh.print_output(res)


def post_commands(ssh, app, socket):
    time.sleep(10)
    ssh.tunnel_login_node(socket=socket, app=app)
    if app.post_command:
        logger.info("Found post command %s" % app.post_command)
        post = app.post_command.replace("$socket_dir", os.path.dirname(socket))
        ssh.execute(shlex.split(post), stream=True)
    time.sleep(30)
    ssh.tunnel_login_node(socket=socket, app=app)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"


schema_url = "https://json-schema.org/draft-07/schema/#"

# App schema and properties

arguments = {
    "type": "object",
    "required": ["name", "description"],
    "properties": {
        "name": {"type": "string"},
        "description": {"type": "string"},
        "split": {"type": ["null", "string"]},
    },
}

appSettingsProperties = {
    "launcher": {"type": "string"},
    "launchers_supported": {"type": "array", "items": {"type": "string"}},
    "script": {"type": "string"},
    "description": {"type": "string"},
    "args": {"type": "array", "items": arguments},
    "examples": {"type": ["array", "string"]},
    "commands": {
        "type": "object",
        "properties": {
            "post": {"type": "string"},
        },
    },
    "needs": {
        "type": "object",
        "properties": {
            "modules": {"type": "array", "items": {"type": "string"}},
            "socket": {"type": "boolean"},
            "xserver": {"type": "boolean"},
        },
    },
}

app_schema = {
    "$schema": schema_url,
    "title": "Tunel App Schema",
    "type": "object",
    "required": ["launcher", "script", "description"],
    "properties": appSettingsProperties,
    "additionalProperties": False,
}

# Launchers properties

singularity_launcher = {
    "type": "object",
    "required": [
        "paths",
    ],
    "properties": {
        "paths": {"type": "array", "items": {"type": "string"}},
        "environment": {"type": "array", "items": {"type": "string"}},
    },
    "additionalProperties": False,
}

# This is also shared by htcondor
slurm_launcher = {
    "type": "object",
    "required": [
        "memory",
        "time",
    ],
    "properties": {
        "memory": {"type": ["string", "number"]},
        "max_attempts": {"type": ["number", "null"]},
        "time": {"type": "string"},
        "paths": {"type": "array", "items": {"type": "string"}},
    },
    "additionalProperties": False,
}


launchers = {
    "type": "object",
    "required": [
        "htcondor",
        "singularity",
        "slurm",
    ],
    "properties": {
        "singularity": singularity_launcher,
        "slurm": slurm_launcher,
        "htcondor": slurm_launcher,
    },
    "additionalProperties": False,
}

shells = ["/bin/bash", "/bin/sh", "/bin/csh"]

# Currently all of these are required
settingsProperties = {
    "ssh_port": {"type": "number"},
    "launchers": launchers,
    "local_port": {"type": "number"},
    "remote_port": {"type": ["number", "null"]},
    "tunel_home": {"type": "string"},
    "tunel_remote_home": {"type": "string"},
    "tunel_remote_work": {"type": "string"},
    "tunel_remote_sockets": {"type": ["string", "null"]},
    "ssh_config": {"type": "string"},
    "ssh_sockets": {"type": "string"},
    "min_port": {"type": "number"},
    "max_port": {"type": "number"},
    "shell": {"type": "string", "enum": shells},
    "config_editor": {"type": "string", "enum": ["nano", "vim", "emacs", "atom"]},
    "apps_dirs": {"type": "array", "items": {"type": "string"}},
    "tunel_spinner": {"type": "string"},
}

settings = {
    "$schema": schema_url,
    "title": "Tunel Settings Schema",
    "type": "object",
    "required": [
        "apps_dirs",
        "local_port",
        "remote_port",
        "ssh_config",
        "ssh_sockets",
        "tunel_home",
        "tunel_remote_home",
        "min_port",
        "max_port",
        "config_editor",
    ],
    "properties": settingsProperties,
    "additionalProperties": False,
}



# This is an example of talking to an API deployed via socket to tunel.
# This uses the tunel-django app, which has a simple joke (JSON) api.
# Usage:
# python tunel_api.py /tmp/test/tunel-django.sock.api.sock

import json
import os
import sys
import tunel.api


def api_get(socket, path="/api/joke/"):
    """
    Get a joke from the tunel-django API.
    This is also provided from the command line in tunel:
    # tunel api-get --socket /tmp/test/tunel-django.sock.api.sock /api/joke/
    """
    api = tunel.api.ApiConnection(socket)
    res = api.get(path)
    print(json.dumps(res.json(), indent=4))
    # print(res.text)


if __name__ == "__main__":
    if len(sys.argv) == 1:
        sys.exit("Please provide the path to a tunel socket.")
    api_get(sys.argv[1])



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"


from tunel.ssh import Tunnel


def main(args, parser, extra, subparser):
    tunel = Tunnel(args.server)
    output = tunel.execute_or_fail(extra)
    for line in output.split("\n"):
        if not line:
            continue
        print(" " + line)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from rich.table import Table

import tunel.apps
from tunel.logger import logger


def list_apps(args, parser, extra, subparser):
    """
    List installed apps
    """
    apps = tunel.apps.list_apps(validate=args.validate)
    if not apps:
        logger.exit("There aren't any applications found in your apps paths.", 0)

    table = Table(title="Tunel Apps")
    table.add_column("#", style="cyan", no_wrap=True)
    table.add_column("Name", style="magenta")
    table.add_column("Launcher", style="magenta")
    table.add_column("Supported", justify="right", style="green")

    for i, app in enumerate(apps.items()):
        name, app = app
        table.add_row(str(i), name, app.launcher, "|".join(app.launchers))

    logger.c.print(table, justify="center")


def docgen(args, parser, extra, subparser):
    """
    Generate docs for apps (to an output directory)
    """
    apps = tunel.apps.list_apps()
    if not apps:
        logger.exit("There aren't any applications found in your apps paths.", 0)

    for i, app in enumerate(apps.items()):
        name, app = app
        logger.info("Generating documentation markdown for %s" % name)
        app.docgen(outdir=args.outdir)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


import os

from jinja2 import Environment, FileSystemLoader

here = os.path.dirname(os.path.abspath(__file__))

# Allow includes from this directory OR providing strings
shared_templates_dir = os.path.join(here, "templates")


class Template:
    """
    Supporting functions for loading any kind of apps template
    """

    def get(self, template_name, template_dir=None):
        """
        Get a template from templates
        """
        template_file = os.path.join(here, "templates", template_name)
        if template_dir and not os.path.exists(template_file):
            template_file = os.path.join(template_dir, template_name)
        if not os.path.exists(template_file):
            template_file = os.path.abspath(template_name)
        return template_file

    def load(self, template_name, template_dir=None):
        """
        Load the default module template.
        """
        template_file = self.get(template_name, template_dir)
        with open(template_file, "r") as temp:
            # If we are given an app template directory, respect it and add to loader
            if template_dir:
                env = Environment(
                    loader=FileSystemLoader([shared_templates_dir, template_dir])
                )
            else:
                env = Environment(loader=FileSystemLoader(shared_templates_dir))
            template = env.from_string(temp.read())
        return template



#!/usr/bin/env python

__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import argparse
import os
import sys

import tunel
from tunel.logger import setup_logger


def get_parser():
    parser = argparse.ArgumentParser(
        description="Tunel for ssh interface tunnels",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Global Variables
    parser.add_argument(
        "--debug",
        dest="debug",
        help="use verbose logging to debug.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--quiet",
        dest="quiet",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--settings-file",
        dest="settings_file",
        help="custom path to settings file.",
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="show software version.",
        default=False,
        action="store_true",
    )

    description = "actions"
    subparsers = parser.add_subparsers(
        help="actions",
        title="actions",
        description=description,
        dest="command",
    )

    # print version and exit
    subparsers.add_parser("version", description="show software version")

    info = subparsers.add_parser(
        "info",
        description="Get information about a named app.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Tunel api-get means we issue a GET request to an endpoint
    api_get = subparsers.add_parser(
        "api-get",
        description="interact with the GET api of a tunel app.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    api_get.add_argument("--socket", help="path to tunel mapped socket")
    api_get.add_argument("path", help="path to GET (defaults to /)")

    for command in info, api_get:
        command.add_argument(
            "--json",
            help="provide result as json",
            default=False,
            action="store_true",
        )

    # Local shell with client loaded
    shell = subparsers.add_parser(
        "shell",
        description="shell into a remote connection",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Also support sh
    sh = subparsers.add_parser(
        "sh",
        description="shell into a remote connection",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Run commands will issue launcher specific commands
    # tunel run-singularity <server> <container>
    run_singularity = subparsers.add_parser(
        "run-singularity",
        description="launch a singularity container.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    run_docker = subparsers.add_parser(
        "run-docker",
        description="launch an interactive docker container.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    run_slurm = subparsers.add_parser(
        "run-slurm",
        description="launch a slurm job or session.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    run_condor = subparsers.add_parser(
        "run-condor",
        description="launch an htcondor job or session.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Stop a slurm running application
    stop_slurm = subparsers.add_parser(
        "stop-slurm",
        description="stop or kill a slurm job or session.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    stop_condor = subparsers.add_parser(
        "stop-condor",
        description="stop or kill an HTCondor job or session.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Apps
    run_app = subparsers.add_parser(
        "run-app",
        description="run a named application",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    stop_app = subparsers.add_parser(
        "stop-app",
        description="stop a named application",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    list_app = subparsers.add_parser(
        "list-apps",
        description="list apps found on the known settings paths",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    list_app.add_argument(
        "--validate",
        help="validate apps loading",
        default=False,
        action="store_true",
    )
    docgen = subparsers.add_parser(
        "docgen",
        description="generate docs (markdown) for apps",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    docgen.add_argument(
        "outdir",
        help="Output directory to write apps markdown",
    )

    # Issue a command to a server
    execute = subparsers.add_parser(
        "exec",
        description="execute a command to a remote connection.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Open up a tunel (from local machine) to remote port
    tunnel = subparsers.add_parser(
        "tunnel",
        description="open up a tunnel to a remote port",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    tunnel.add_argument(
        "--isolated-nodes",
        help="if on a cluster, nodes are isolated",
        default=False,
        action="store_true",
    )

    launchers = [
        run_docker,
        run_singularity,
        run_slurm,
        run_app,
        stop_app,
        stop_slurm,
        run_condor,
        stop_condor,
    ]

    # Run app can take a --launcher preference
    for command in run_app, stop_app:
        command.add_argument("--launcher", help="specify a non-default launcher to use")

    for command in [tunnel] + launchers:
        command.add_argument("--port", help="remote port to connect to.")
        command.add_argument("--local-port", help="local port to connect to.")

    for command in [
        sh,
        shell,
        execute,
        tunnel,
    ] + launchers:
        command.add_argument(
            "server",
            help="server identity to interact with (e.g., name in ~/.ssh/config)",
        )

    for command in [run_app, stop_app, info]:
        command.add_argument("app", help="The name of the application.")
    return parser


def run_tunel():

    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client
        and exit with return code.
        """

        version = tunel.__version__

        print("\nTunel v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    # TODO this could eventually run a UI
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    if args.debug is True:
        os.environ["MESSAGELEVEL"] = "DEBUG"

    # Show the version and exit
    if args.command == "version" or args.version:
        print(tunel.__version__)
        sys.exit(0)

    setup_logger(
        quiet=args.quiet,
        debug=args.debug,
    )

    # retrieve subparser (with help) from parser
    helper = None
    subparsers_actions = [
        action
        for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)
    ]
    for subparsers_action in subparsers_actions:
        for choice, subparser in subparsers_action.choices.items():
            if choice == args.command:
                helper = subparser
                break

    # Does the user want a shell?
    if args.command in ["shell", "sh"]:
        from .shell import main
    if args.command == "api-get":
        from .api import api_get as main
    if args.command == "docgen":
        from .apps import docgen as main
    if args.command == "exec":
        from .execute import main
    if args.command == "info":
        from .info import main
    if args.command == "tunnel":
        from .tunnel import main
    if args.command == "run-docker":
        from .launcher import run_docker as main
    if args.command == "run-singularity":
        from .launcher import run_singularity as main
    if args.command == "run-slurm":
        from .launcher import run_slurm as main
    if args.command == "run-condor":
        from .launcher import run_condor as main
    if args.command == "stop-slurm":
        from .launcher import stop_slurm as main
    if args.command == "stop-condor":
        from .launcher import stop_condor as main
    if args.command == "run-app":
        from .launcher import run_app as main
    if args.command == "stop-app":
        from .launcher import stop_app as main
    if args.command == "list-apps":
        from .apps import list_apps as main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args=args, parser=parser, extra=extra, subparser=helper)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)


if __name__ == "__main__":
    run_tunel()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os

import tunel.utils as utils

install_dir = utils.get_installdir()

# Default apps directory
default_apps_dir = os.path.join(install_dir, "apps")

reps = {
    "$install_dir": install_dir,
    "$root_dir": os.path.dirname(install_dir),
    "$default_apps": default_apps_dir,
}

# The default settings file in the install root
default_settings_file = os.path.join(reps["$install_dir"], "settings.yml")

# The user settings file can be created to over-ride default
user_settings_file = os.path.join(os.path.expanduser("~/.tunel"), "settings.yml")

# variables in config we allow environment substitution
allowed_envars = ["ssh_config", "ssh_sockets", "tunel_home"]

# The GitHub repository with recipes (apps)
github_url = "https://github.com/tunel-apps/tunel"



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os
import sys
from datetime import datetime

import jsonschema

import tunel.defaults as defaults
import tunel.schemas
import tunel.template
import tunel.utils as utils
from tunel.logger import logger
from tunel.template import Template

from .settings import Settings


class App:
    def __init__(self, app_config, app_dir, validate=True):
        self.args = {}
        self.app_config = app_config
        self.app_root = app_dir
        self.load(validate)

    def __str__(self):
        return "[%s:%s]" % (self.launcher, self.name)

    def __repr__(self):
        return self.__str__()

    @property
    def app_dir(self):
        return os.path.dirname(self.app_config)

    @property
    def name(self):
        return os.path.dirname(self.relative_path)

    @property
    def job_name(self):
        return self.name.replace(os.sep, "-")

    @property
    def relative_path(self):
        return self.app_config.replace(self.app_root, "").strip("/")

    @property
    def relative_dir(self):
        return os.path.dirname(self.relative_path)

    def get(self, key, default=None):
        return self.config.get(key, default)

    @property
    def launchers(self):
        if not self.launchers_supported:
            return [self.launcher]
        return list(set(self.launchers_supported + [self.launcher]))

    @property
    def has_xserver(self):
        """
        Boolean to indicate if an application has an xserver
        """
        if self.needs and self.needs.get("xserver", False):
            return True
        return False

    def get_script(self):
        return os.path.join(self.app_dir, self.script)

    def __getattr__(self, key):
        """
        A direct get of an attribute, but default to None if doesn't exist
        """
        return self.get(key)

    def print_pretty(self):
        """
        Print the app (and args, etc) pretty to the terminal.
        """
        logger.print_pretty(self.config)

    def docgen(self, out=None, outdir=None):
        """
        Render documentation for an app.
        """
        if outdir and not os.path.exists(outdir):
            logger.exit(f"{outdir} does not exist.")
        if outdir and os.path.exists(outdir):
            out = os.path.join(outdir, "%s.md" % self.job_name)
            out = open(out, "w")
        out = out or sys.stdout
        template = tunel.template.Template().load("docs.md")

        github_url = "%s/blob/main/tunel/apps/%s" % (
            defaults.github_url,
            self.relative_path,
        )
        script_url = "%s/blob/main/tunel/apps/%s/%s" % (
            defaults.github_url,
            self.relative_dir,
            self.script,
        )
        script = utils.read_file(self.get_script())

        # Currently one doc is rendered for all containers
        result = template.render(
            app=self,
            script="{% raw %}" + script + "{% endraw %}",
            script_url=script_url,
            github_url=github_url,
            creation_date=datetime.now(),
        )
        out.write(result)
        out.close()
        return result

    def add_args(self, extras):
        """
        Extra is a list of args, in either format:

        --name=value
        --flag

        They must start with -- to be considered.
        """
        args = {x["name"]: x.get("split") for x in self.config.get("args", {})}
        argkeys = ", ".join(list(args.keys()))
        for extra in extras:
            if not extra.startswith("--"):
                continue
            arg = extra.lstrip("--").split("=")
            if arg[0] not in args:
                logger.warning(
                    f"{arg} is not a recognized argument, choices are {argkeys}"
                )
            if len(arg) == 1:
                self.args[arg[0]] = True
            elif len(arg) == 2:
                value = arg[1]
                splitby = args.get(arg[0])
                if splitby:
                    value = value.split(splitby)
                self.args[arg[0]] = value
            else:
                logger.warning(f"Unexpected argument: {extra}, skipping.")

    @property
    def post_command(self):
        """
        Get a post command, if it exists
        """
        commands = self.get("commands")
        if not commands:
            return
        return commands.get("post")

    def load_template(self):
        """
        Given an app, load the template script for it.

        This also provides the app directory for any relative template scripts.
        """
        template = Template()
        return template.load(self.get_script(), self.app_dir)

    def validate(self):
        """
        Validate the loaded settings with jsonschema
        """
        jsonschema.validate(self.config, schema=tunel.schemas.app_schema)

    def load(self, validate=True):
        self.config = utils.read_yaml(self.app_config)
        if validate:
            self.validate()


def get_app(name, extra=None):
    """
    Given an app name, get the loaded app for it
    """
    apps = []
    listing = list_apps()
    for app in listing:
        if name in app:
            apps.append(app)

    if not apps:
        logger.exit(
            "Cannot find app %s\nChoices are:\n%s" % (name, "\n  ".join(listing))
        )
    if len(apps) > 1:
        logger.exit(
            "Found %s apps:\n%s\nBe more specific to disambiguate %s!"
            % (len(apps), "\n".join(apps), name)
        )

    app = apps[0]
    logger.info(f"Loading app {app}...")
    app = listing[app]
    app.add_args(extra)
    return app


def list_apps(settings_file=None, validate=False):
    """
    Given the user settings,
    """
    settings_file = settings_file or defaults.default_settings_file
    settings = Settings(settings_file)
    apps = {}

    # parse through in reverse so top names take priority
    for app_dir in reversed(settings.apps_dirs):
        if not os.path.exists(app_dir):
            logger.warning("%s does not exist." % app_dir)
            continue

        # Look for apps that validate
        for filename in utils.recursive_find(app_dir, "app.yaml"):
            try:
                app = App(filename, app_dir)
            except Exception as e:
                if validate:
                    logger.exit("%s is not valid: %s" % (filename, e))
                else:
                    logger.error("%s is not valid" % (filename))
                    continue
            apps[app.name] = app

    return apps



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"


import os
import random
import shlex

import tunel.defaults as defaults
import tunel.ssh.commands as commands
import tunel.utils
from tunel.logger import logger
from tunel.settings import Settings


class Tunnel:
    """
    An ssh tunnel provides a route for the user to interact with an application
    This basically enables local port forwarding (called an ss tunnel) using
    the library Paramiko.
    """

    def __init__(self, server, **kwargs):

        # If/when we open a shell
        self.ssh = None
        self.username = None
        self.settings_file = (
            kwargs.get("settings_file") or defaults.default_settings_file
        )
        self.settings = Settings(self.settings_file)
        self.local_port = int(kwargs.get("local_port") or self.settings.local_port)
        self.remote_port = int(
            kwargs.get("remote_port")
            or self.settings.remote_port
            or self._random_port()
        )

        # Local port to forward to
        self.web_port = kwargs.get("web_port") or self.settings.web_port

        # Save the name of the server to connect to
        self.server = server

    def _random_port(self):
        """
        Generate a random port for an ssh session
        """
        return random.choice(range(self.settings.min_port, self.settings.max_port))

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[tunel-ssh]"

    def execute(self, cmd, login_shell=False, quiet=False, stream=False, xserver=False):
        """
        Execute a command via ssh and a known, named connection.
        """
        if not isinstance(cmd, list):
            cmd = shlex.split(cmd)
        if stream:
            command = ["ssh", "-t", self.server]
        elif xserver:
            command = ["ssh", "-XY", "-o", "ForwardX11=yes", self.server]
        else:
            command = ["ssh", self.server]

        cmd = command + cmd
        return tunel.utils.run_command(cmd, quiet=quiet, stream=stream)

    def scp_to(self, src, dest):
        """
        Copy a file onto the server
        """
        self.execute_or_fail("mkdir -p %s" % os.path.dirname(dest))
        cmd = ["scp", src, "%s:%s" % (self.server, dest)]
        return tunel.utils.run_command(cmd)

    def scp_from(self, src, dest):
        """
        Copy a file from the server to local
        """
        cmd = ["scp", "%s:%s" % (self.server, src), dest]
        return tunel.utils.run_command(cmd)

    def print_output(self, output, success_code=0, quiet=False):
        """
        Given an output dict, print and color appropriately.
        """
        if output["return_code"] != success_code and output["message"]:
            logger.error(output["message"].strip())
        elif not quiet and output["message"]:
            logger.info(output["message"].strip())

    def execute_or_fail(self, cmd, success_code=0, quiet=False):
        """
        Execute a command, show the command preview, only continue on success.
        """
        output = self.execute(cmd, quiet=quiet)
        if output["return_code"] != success_code:
            logger.exit(output["message"])
        return output["message"].strip()

    def tunnel_login_node(self, port=None, remote_port=None, socket=None, app=None):
        """
        Create a tunnel directly to the login node (e.g., a Singularity app)
        """
        # The app requires a socket
        if app.needs and app.needs.get("socket", False) is True:
            if not socket:
                logger.exit("A socket path is required.")
            return self._tunnel_login_node_socket(socket)

        # The app uses an xserver
        elif app.needs and app.needs.get("xserver", False) is True:
            return self._tunnel_login_node_xserver()

        port = port or self.local_port
        remote_port = remote_port or self.remote_port
        logger.info(
            "Forwarding port %s to %s:%s ..." % (port, self.server, remote_port)
        )
        return self._tunnel_login_node_port()

    def tunnel(
        self,
        machine=None,
        port=None,
        remote_port=None,
        socket=None,
        app=None,
    ):
        """
        Given a remote and local port, open a tunnel. If an isolated node ssh is
        done, the name of the machine is required too.
        """
        # If no machine, we have to do a login node
        if not machine:
            return self._tunnel_login()

        # The app requires a socket
        if app.needs.get("socket", False):
            if not socket:
                logger.exit("A socket path is required.")
            return self._tunnel_isolated_socket(machine, socket=socket)

        port = port or self.local_port
        remote_port = remote_port or self.remote_port
        logger.info(
            "Forwarding port %s to %s:%s ..." % (port, self.server, remote_port)
        )
        return self._tunnel_isolated_port(machine)

    def _get_socket_path(self):
        """
        Get a path for the socket to control the connection. Should be in ~/.ssh
        """
        socket_dir = self.settings.ssh_sockets
        if not os.path.exists(socket_dir):
            os.mkdir(socket_dir)
        return tunel.utils.get_tmpfile(socket_dir, prefix=self.server)

    def _close_socket(self, socket_file):
        """
        Ensure an ssh socket is closed
        """
        cmd = ["-S", socket_file, "-O", "exit", self.server]
        try:
            self.execute(cmd)
        except:
            pass
        if os.path.exists(socket_file):
            os.remove(socket_file)

    def shell(self, cmd=None, interactive=False):
        """
        Pass the process over to shell
        """
        command = "ssh %s" % self.server
        if interactive:
            command = "ssh -t %s" % self.server
        if cmd:
            command = "%s %s" % (command, cmd)
        os.system(command)
        logger.info("👋️ Goodbye!")


# Add tunnel commands to class (only done for organization)

Tunnel._tunnel_wait = commands._tunnel_wait
Tunnel._tunnel_isolated_socket = commands._tunnel_isolated_socket
Tunnel._tunnel_isolated_sockets = commands._tunnel_isolated_sockets
Tunnel._tunnel_isolated_proxyjump_sockets = commands._tunnel_isolated_proxyjump_sockets
Tunnel._tunnel_isolated_port = commands._tunnel_isolated_port
Tunnel._tunnel_login = commands._tunnel_login
Tunnel._tunnel_login_node_port = commands._tunnel_login_node_port
Tunnel._tunnel_login_node_socket = commands._tunnel_login_node_socket
Tunnel._tunnel_login_node_xserver = commands._tunnel_login_node_xserver



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

# Original example used to derive this is provided under the MIT License
# https://github.com/akx/uwsgi-socket-example/commit/e5e1df3cdc4a860434b7655663133d8e5103e431

import json
import os
import socket
import struct
import sys


def force_bytes(value):
    """
    Force always returning bytes
    """
    return value if isinstance(value, bytes) else str(value).encode("utf-8")


class ApiResponse:
    def __init__(self, url, raw):
        self.raw = raw
        self.headers = {}
        self.text = ""
        self.parse()

    def parse(self):
        """
        Parse the raw response.
        """
        # headers and content split by two lines
        # Can we use a proper library here?
        headers, content = self.raw.split("\r\n\r\n")
        for header in headers:
            if not header or ":" not in header:
                continue
            k, v = header.split(":", 1)
            self.headers[k.strip()] = v.strip()

        # Raw text
        self.text = content

    def json(self):
        return json.loads(self.text)


class ApiConnection:
    def __init__(self, path, server_name="0.0.0.0", server_port="8000"):
        """
        Create a new tunel API connection.

        Server name and port only matter if your server implements rate limiting
        or other logic to determine if you should be allowed/blocked.
        """
        self.s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.path = path
        self.server_name = server_name
        self.server_port = server_port
        if not os.path.exists(path):
            sys.exit("%s does not exist." % path)

    def connect(self):
        self.s.connect(self.path)

    def __exit__(self):
        self.disconnect()

    def disconnect(self):
        self.s.close()

    def send_uwsgi_request(self, header_content):
        """
        Prepare request headers and data, and send.
        """
        data = encode_uwsgi_vars(header_content)
        header = struct.pack(
            "<BHB",
            0,  # modifier1: 0 - WSGI (Python) request
            len(data),  # data size
            0,  # modifier2: 0 - always zero
        )
        self.s.sendall(header)
        self.s.sendall(data)

    def get_response(self, url, width=32):
        """
        Dump the response from the socket.
        """
        # First assemble the content into a string (assumes smaller)
        content = ""
        while True:
            chunk = self.s.recv(width)
            if not chunk:
                break
            content += chunk.decode("ascii", "replace")
        return ApiResponse(url, content)

    def get(self, path):
        """
        Akin to a GET request for a particular path.
        """
        self.connect()
        self.send_uwsgi_request(
            {
                "PATH_INFO": path,
                "REQUEST_METHOD": "GET",
                "SERVER_NAME": self.server_name,
                "SERVER_PORT": self.server_port,
            }.items()
        )
        result = self.get_response(path)
        self.disconnect()
        return result


def encode_uwsgi_vars(values):
    """
    Encode a list of key-value pairs into an uWSGI request header structure.
    """
    # See http://uwsgi-docs.readthedocs.io/en/latest/Protocol.html#the-uwsgi-vars
    buf = []
    for key, value in values:
        key_enc = force_bytes(key)
        val_enc = force_bytes(value)
        buf.append(struct.pack("<H", len(key_enc)))
        buf.append(key_enc)
        buf.append(struct.pack("<H", len(val_enc)))
        buf.append(val_enc)
    return b"".join(buf)



from .fileio import (
    copyfile,
    get_file_hash,
    get_tmpdir,
    get_tmpfile,
    mkdir_p,
    mkdirp,
    print_json,
    read_config_file,
    read_file,
    read_json,
    read_lines,
    read_yaml,
    recursive_find,
    write_file,
    write_json,
)
from .generate import namer
from .terminal import (
    check_install,
    confirm_action,
    confirm_uninstall,
    get_installdir,
    run_command,
    which,
)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"


import tunel.apps
import tunel.launcher

# Generic apps


def run_app(args, parser, extra, subparser):
    app = tunel.apps.get_app(args.app, extra)
    launcher = tunel.launcher.get_launcher(app, args.launcher)(
        args.server, remote_port=args.port, local_port=args.local_port
    )
    launcher.run_app(app)


def stop_app(args, parser, extra, subparser):
    app = tunel.apps.get_app(args.app, extra)
    launcher = tunel.launcher.get_launcher(app, args.launcher)(
        args.server, remote_port=args.port, local_port=args.local_port
    )
    launcher.stop_app(app)


# Specific to launcher


def run_condor(args, parser, extra, subparser):

    launcher = tunel.launcher.HTCondor(
        args.server, remote_port=args.port, local_port=args.local_port
    )
    launcher.run(extra)


def stop_condor(args, parser, extra, subparser):
    launcher = tunel.launcher.HTCondor(
        args.server, remote_port=args.port, local_port=args.local_port
    )
    launcher.stop(extra)


def run_singularity(args, parser, extra, subparser):
    launcher = tunel.launcher.Singularity(
        args.server, remote_port=args.port, local_port=args.local_port
    )
    launcher.run(extra)


def run_docker(args, parser, extra, subparser):
    launcher = tunel.launcher.Docker(
        args.server, remote_port=args.port, local_port=args.local_port
    )
    launcher.run(extra)


def stop_slurm(args, parser, extra, subparser):
    launcher = tunel.launcher.Slurm(
        args.server, remote_port=args.port, local_port=args.local_port
    )
    launcher.stop(extra)


def run_slurm(args, parser, extra, subparser):

    launcher = tunel.launcher.Slurm(
        args.server, remote_port=args.port, local_port=args.local_port
    )
    launcher.run(extra)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import signal

from tunel.logger import logger

# Shared tunnel functions


def _tunnel_wait(self, socket_file=None):
    """
    Wait for a Control+C to exit and remove a tunnel
    """

    def signal_handler(sig, frame):
        if socket_file:
            self._close_socket(socket_file)
        logger.exit("🛑️ Port forwarding stopped.", return_code=0)

    signal.signal(signal.SIGINT, signal_handler)
    print("Press Ctrl+C")
    signal.pause()


# Tunnel to remote nodes


def _tunnel_isolated_socket(self, machine, socket):
    """
    show commands for 1. isolated socket, and 2. proxyjump.
    """
    self._tunnel_isolated_sockets(machine, socket)
    self._tunnel_isolated_proxyjump_sockets(machine, socket)


def _tunnel_isolated_sockets(self, machine, socket):
    """
    This approach maps a socket first to the login node and connects to it.

    # Running on login node
    $ ssh -NT -L 8888:/tmp/test.sock user@server

    # TWO COMMANDS (and assuming isolated node)
    $ ssh -NT user@server ssh <machine> -NT -L /home/user/login-node.sock:/home/user/path/to/worker-node.sock

    # And another for the local socket
    $ ssh -NT -L <localport>:/home/user/login-node.sock user@server

    # ONE COMMAND
    $ ssh -NT -L <localport>:/home/user/login-node.sock user@server ssh <machine> -NT -L /home/user/login.node.sock:/home/user/path/to/worker-node.sock
    """
    login_node_socket = socket.replace(".sock", ".head-node.sock")
    cmd = [
        "ssh",
        "-NT",
        "-L",
        "%s:%s" % (self.local_port, login_node_socket),
        "%s@%s" % (self.username, self.server),
        "ssh",
        "%s@%s" % (self.username, machine),
        "-NT",
        "-L",
        "%s:%s" % (login_node_socket, socket),
    ]
    logger.c.print()
    logger.c.print("== RUN THIS IN A SEPARATE TERMINAL AFTER THE APP IS READY ==")
    logger.info("%s" % " ".join(cmd))


def _tunnel_isolated_proxyjump_sockets(self, machine, socket):
    """
    This approach uses a proxyjump to only require one socket

    # OR us a proxy
    ssh -J user@server <machine> -NT -L <port>:/home/user/path/to/worker-node.sock
    """
    logger.c.print("== OR (newer ssh) USE A PROXYJUMP ==")
    cmd = [
        "ssh",
        "-J",
        "%s@%s" % (self.username, self.server),
        "%s@%s" % (self.username, machine),
        "-NT",
        "-L",
        "%s:%s" % (self.local_port, socket),
    ]
    logger.info("%s" % " ".join(cmd))


def _tunnel_isolated_port(self, machine):
    """
    Create a tunnel to an isolated node (not tested yet)
    """
    # TODO do we need to close up connections on login node?
    connection = "%s:localhost:%s" % (self.local_port, self.remote_port)
    cmd = [
        "-f",
        "-L",
        connection,
        self.server,
        "ssh",
        "-L",
        connection,
        "-N",
        machine,
    ]
    self.execute(cmd)
    self._tunnel_wait()


# Tunnels to login node


def _tunnel_login(self):
    """
    Create a simple tunnel to the login node (assumes not isolated nodes)
    """
    socket_file = self._get_socket_path()
    cmd = ["-K", "-f", "-M"]

    # Add the socket file
    cmd += ["-S", socket_file, "-L"]
    cmd += ["%s:%s:%s" % (self.local_port, self.server, self.remote_port), "-N"]
    self.execute(cmd)
    self._tunnel_wait(socket_file)


def _tunnel_login_node_socket(self, socket):
    """
    Connect to a login node socket
    """
    cmd = [
        "ssh",
        "-NT",
        "-L",
        "%s:%s" % (self.local_port, socket),
        "%s@%s" % (self.username, self.server),
    ]
    cmd = "%s" % " ".join(cmd)
    logger.info(cmd)


def _tunnel_login_node_xserver(self):
    """
    Connect to a login node via xserver
    """
    cmd = [
        "ssh",
        "-o",
        "ForwardX11=yes",
        "-X",
        "-NT",
        "-L",
        "%s@%s" % (self.username, self.server),
    ]
    cmd = "%s" % " ".join(cmd)
    logger.info(cmd)


def _tunnel_login_node_port(self, machine):
    """
    Create a tunnel to login node port
    TODO this needs testing.
    """
    socket_path = self._get_socket_path()

    # TODO do we need to close up connections on login node?
    connection = "%s:localhost:%s" % (self.local_port, self.remote_port)
    cmd = [
        "-f",
        "-S",
        socket_path,
        "-L",
        connection,
        self.server,
    ]
    self.execute(cmd)
    self._tunnel_wait(socket_path)



from tunel.logger import logger

from .base import Launcher
from .docker import Docker
from .htcondor import HTCondor
from .podman import Podman
from .singularity import Singularity
from .slurm import Slurm


def get_launcher(app, launcher=None):
    """
    Get a launcher for a loaded app
    """
    # No go if the launcher is defined and not tested for the app
    if launcher and launcher not in app.launchers:
        logger.exit(f"Launcher {launcher} has not been tested for app {app.name}")

    # Otherwise, should be ok!
    if launcher:
        app.launcher = launcher

    if app.launcher in ["htcondor", "condor"]:
        return HTCondor
    if app.launcher == "slurm":
        return Slurm
    if app.launcher == "podman":
        return Podman
    if app.launcher == "docker":
        return Docker
    elif app.launcher == "singularity":
        return Singularity
    return Launcher



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os

import tunel.ssh
import tunel.template
from tunel.launcher.base import Launcher
from tunel.logger import logger

here = os.path.dirname(os.path.abspath(__file__))


class HTCondor(Launcher):
    """
    An HTCondor interacts with condor
    """

    def run_app(self, app):
        """
        Given an app designated for htcondor, run it!
        """
        # Add any paths from the config
        paths = self.settings.get("paths", [])

        # Prepare dictionary with content to render into recipes
        render = self.prepare_render(app, paths)

        # Update memory if not in args
        if "memory" not in render["args"]:
            render["args"]["memory"] = self.settings["memory"]

        # Clean up previous sockets
        self.ssh.execute(["rm", "-rf", "%s/*.sock" % render["scriptdir"]])

        # Load the app template
        template = app.load_template()
        result = template.render(**render)

        # We also need to render the template for the submit
        template = tunel.template.Template()
        job_template = template.load("htcondor/template.submit")
        job_submit = job_template.render(**render)

        # Write the job script to the cluster
        tmpfile = self.write_temporary_script(result)
        self.ssh.scp_to(tmpfile, render["script"])
        os.remove(tmpfile)

        # And the submit script
        tmpfile = self.write_temporary_script(job_submit)
        submit_file = render["script"] + ".submit"
        self.ssh.scp_to(tmpfile, submit_file)
        os.remove(tmpfile)

        # Assemble the command
        command = ["condor_submit", submit_file, "-batch-name", app.job_name]

        # Launch with command
        if not self.previous_job_exists(app.job_name):
            self.run(command)

    def stop_app(self, app):
        """
        Wrapper to stop a single app.
        """
        return self.stop([app.name])

    def stop(self, names):
        """
        Stop one or more named jobs
        """
        for name in names:
            logger.purple("Killing %s condor job on %s" % (name, self.ssh.server))
            job = self.get_job(name)
            if job:
                _, jobid = job.rsplit(" ", 1)
                self.ssh.execute_or_fail("condor_rm %s" % jobid)

    def get_job(self, name):
        """
        Get a job by name (the unparsed line)
        """
        lines = self.ssh.execute_or_fail("condor_q %s" % self.username)
        for line in lines.split("\n"):
            if line.startswith(self.username):
                owner, job_name, rest = line.split(" ", 2)
                if name == job_name:
                    return line

    def previous_job_exists(self, name):
        """
        Check to see if a previous job was submit/exists.
        """
        logger.purple("Checking for previous run of %s" % name)
        job = self.get_job(name)
        if job:
            logger.error("Found existing job for %s!" % name)
            logger.exit(
                "Please run 'tunel stop-condor %s %s' before proceeding."
                % (self.ssh.server, name)
            )
        logger.purple("No existing %s jobs found, continuing." % name)

    def run(self, cmd):
        """
        Run a command for htcondor
        """
        # If no command, get interactive node
        if not cmd:
            logger.info("No command supplied, will init interactive session!")
            self.ssh.shell("condor_submit -interactive", interactive=True)

        else:
            res = self.ssh.execute(" ".join(cmd))

            # A successful submission should:
            if res["return_code"] == 0:
                logger.c.print(res["message"])



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from tunel.logger import logger

from .container import ContainerLauncher


class Docker(ContainerLauncher):
    name = "docker"

    def run(self, *args, **kwargs):
        """
        Run handles some command to docker (this can be expanded in functionality)
        """
        if self.path:
            command = "docker run -d --rm -i %s PATH=%s %s" % (
                self.env_command,
                self.path,
                " ".join(args[0]),
            )
        else:
            command = "docker run -d --rm -i %s %s" % (
                self.env_command,
                " ".join(args[0]),
            )

        res = self.ssh.execute(command)
        self.ssh.print_output(res)

    @property
    def env_command(self):
        """
        Transform a listing of key=value into docker --env flags
        """
        envars = self.environ.split(" ") if self.environ else ""
        if envars:
            envars = ["--env %s" for e in envars]
            envars = " ".join(envars)
        return envars

    def stop_app(self, app):
        """
        Wrapper to stop a single app.
        """
        return self.stop([app.name])

    def stop(self, names):
        """
        Stop one or more named jobs
        """
        for name in names:
            name = name.replace("/", "-")
            logger.purple("Killing %s docker container on %s" % (name, self.ssh.server))
            self.ssh.execute_or_fail("docker stop %s" % name)
            self.ssh.execute_or_fail("docker rm %s || echo Already removed." % name)
